-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

local ILVL_ACTIVEHEIRLOOM = -1;
local LOWLEVEL_MULTIPLIER = 0.8;
local MAX_PET_LEVEL = 25;
local STRIPE_COLOR = {r=0.9, g=0.9, b=1};

-- Names of the equipment slots
local equipSlotNames =
{
  [INVSLOT_HEAD    ] = "Head",
  [INVSLOT_NECK    ] = "Neck",
  [INVSLOT_SHOULDER] = "Shoulder",
  [INVSLOT_CHEST   ] = "Chest",
  [INVSLOT_WAIST   ] = "Waist",
  [INVSLOT_LEGS    ] = "Legs",
  [INVSLOT_FEET    ] = "Feet",
  [INVSLOT_WRIST   ] = "Wrist",
  [INVSLOT_HAND    ] = "Hands",
  [INVSLOT_FINGER1 ] = "Finger0",
  [INVSLOT_FINGER2 ] = "Finger1",
  [INVSLOT_TRINKET1] = "Trinket0",
  [INVSLOT_TRINKET2] = "Trinket1",
  [INVSLOT_BACK    ] = "Back",
  [INVSLOT_MAINHAND] = "MainHand",
  [INVSLOT_OFFHAND ] = "SecondaryHand",
};

-- Info for equipped items
local equippedItems =
{
  [INVSLOT_HEAD    ] = {id = 0, iLvl = 0},
  [INVSLOT_NECK    ] = {id = 0, iLvl = 0},
  [INVSLOT_SHOULDER] = {id = 0, iLvl = 0},
  [INVSLOT_CHEST   ] = {id = 0, iLvl = 0},
  [INVSLOT_WAIST   ] = {id = 0, iLvl = 0},
  [INVSLOT_LEGS    ] = {id = 0, iLvl = 0},
  [INVSLOT_FEET    ] = {id = 0, iLvl = 0},
  [INVSLOT_WRIST   ] = {id = 0, iLvl = 0},
  [INVSLOT_HAND    ] = {id = 0, iLvl = 0},
  [INVSLOT_FINGER1 ] = {id = 0, iLvl = 0},
  [INVSLOT_FINGER2 ] = {id = 0, iLvl = 0},
  [INVSLOT_TRINKET1] = {id = 0, iLvl = 0},
  [INVSLOT_TRINKET2] = {id = 0, iLvl = 0},
  [INVSLOT_BACK    ] = {id = 0, iLvl = 0},
  [INVSLOT_MAINHAND] = {id = 0, iLvl = 0},
  [INVSLOT_OFFHAND ] = {id = 0, iLvl = 0},
};

-- Stats for equipped slots
local equippedStats =
{
  count   = 0,
  lowIlvl = 0,
  empty   = 0,
};

-- Auction sub-categories for 2H items that aren't actually flagged as 2H
local equippedTwoHanded =
{
  ["Bows"     ] = true,
  ["Crossbows"] = true,
  ["Guns"]      = true,
};

-- Item IDs for all know lances
local tournamentLances =
{
  [46069] = true, -- Alliance Lance
  [46070] = true, -- Horde Lance
  [46106] = true, -- Argent Lance
  [52716] = true  -- Twilight Firelance
};

-- Info for Pet Battle items that grant an XP bonus
local petBattleItems =
{
  {id = 92738, bonus = 10, slot = INVSLOT_HEAD} -- Safari Hat
};

-- Info for the equipment set specialization frames
local equipSetSpecInfo =
{
  dialogHeight = 0,
  textures = {},
  fontStrings = {},
  buttons = {},
  selectedSpec = 0,
  setName = nil
};

local filteredEquipSets = {[0] = 0};

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to build a list of equipment sets marked as being for the active
-- specialization.
--
local function equipSetPopupFilter()
  -- Local Variables
  local activeSpecId;
  local activeSpecIdx = GetSpecialization();
  local equipSetSpec = NigglesDBC.equipSetSpec;
  local numFiltered = 0;
  local setName;
  
  -- Check there is an active spec
  if (activeSpecIdx ~= nil) then
    -- Get the ID of the active spec
    activeSpecId = GetSpecializationInfo(activeSpecIdx);

    -- Filter the equipment sets
    if (activeSpecId ~= nil) then
      for idx = 1, GetNumEquipmentSets() do
        setName = GetEquipmentSetInfo(idx);
        if (equipSetSpec[setName] == activeSpecId) then
          numFiltered = numFiltered+1;
          filteredEquipSets[numFiltered] = idx;
        end
      end
    end
  end

  -- Save the number of filtered equipment sets
  filteredEquipSets[0] = numFiltered;

  -- Check if the selection is no longer valid
  if ((NigglesEquipSetPopup.selectedSet ~= nil) and
      (NigglesDBC.equipSetSpec[NigglesEquipSetPopup.selectedSet] ~=
        activeSpecId)) then
    NigglesEquipSetPopup.selectedSet = nil;
    NigglesEquipSetPopup.equip:Disable();
  end

  -- Update the Equip Set popup
  if (NigglesEquipSetPopup:IsShown()) then
    NigglesEquipSetPopupUpdate();
  end
 
  return;
end

--
-- Function called to process events for the Niggles 'Equip Set' popup
--
local function equipSetPopupOnEvent(self, event, ...)
  -- Process the event
  if (event == "EQUIPMENT_SWAP_FINISHED") then
    local wasSuccessful, setName = ...;
    if ((wasSuccessful) and (setName ~= nil)) then
      NigglesEquipSetEquipped(setName);
    end
  elseif (event == "ACTIVE_TALENT_GROUP_CHANGED") then
    if (self.processEvent) then
      NigglesEquipSetSpecGroupChanged();
    else
      self.processEvent = true;
    end
  elseif (event == "EQUIPMENT_SETS_CHANGED") then
    if (NigglesEquipSetPopup:IsShown()) then
      equipSetPopupFilter();
    end
  elseif (event == "PLAYER_EQUIPMENT_CHANGED") then
    if ((NigglesDBC.settings.UiLowLevelItems) or
        (NigglesDBC.settings.WarningLowLevelItems)) then
      NigglesEquipmentScan();
    end
    if (NigglesEquipSetPopup:IsShown()) then
      equipSetPopupFilter();
    end
  elseif (event == "UNIT_SPELLCAST_INTERRUPTED") then
    local _, _, _, _, spellId = ...;
    if ((spellId == 63645) or    -- 'Activating Primary Spec'
        (spellId == 63644)) then -- 'Activating Secondary Spec'
      self:UnregisterEvent("UNIT_SPELLCAST_INTERRUPTED");
      self.processEvent = true;
    end
  end

  return;
end

--
-- Function called when the Niggles 'Equip Set' popup is shown
--
local function equipSetPopupOnShow(self)
  -- Clear any old selection
  self.equip:Disable();
  self.selectedSet = nil;

  -- Filter the equipment sets
  equipSetPopupFilter();

  return;
end

--
-- Function called when the 'OK' button is clicked in the EM Set dialog
--
local function equipSetSpecOkayOnClick(self, mouseBtn)
  -- Local Variables
  local numSpecs = (UnitLevel("player") >= 10 and GetNumSpecGroups() or 0);

  -- Check if the EM Set spec feature is enabled
  if (NigglesDBC.settings.UiEMSetSpec) then
    -- Local Variables
    local origName = GearManagerDialogPopup.origName;

    -- Check a name has been specified for the EM set
    if (equipSetSpecInfo.setName ~= nil) then
      -- Check if the EM set has been renamed
      if ((origName ~= nil) and
          (origName ~= equipSetSpecInfo.name) and
          (NigglesDBC.equipSetSpec[origName] ~= nil)) then
        -- ...clear its entry
        NigglesDBC.equipSetSpec[origName] = nil;
      end

      -- Check if the player can have specs and has selected one
      if ((numSpecs > 1) and (equipSetSpecInfo.selectedSpec ~= 0)) then
        -- Save the EM set's spec
        NigglesDBC.equipSetSpec[equipSetSpecInfo.setName] =
          equipSetSpecInfo.selectedSpec;
      else
        NigglesDBC.equipSetSpec[equipSetSpecInfo.setName] = nil;
      end
    end

    -- Refresh the 'Equip Set' popup, if required
    if (NigglesEquipSetPopup:IsShown()) then
      equipSetPopupFilter();
    end
  end

  return;
end

--
-- Function to pre-process a click on the 'OK' button in the EM Set dialog.
-- It saves the name of the EM set. This function is required because the
-- 'OnClick' function for the button clears the name from both it's internal
-- variable and the edit box in the dialog.
--
local function equipSetSpecOkayPreClick(self, mouseBtn)
  -- Save the name of the EM set
  equipSetSpecInfo.setName = GearManagerDialogPopup.name;

  return;
end

--
-- Function called when 'RecalculateGearManagerDialogPopup' is called by the
-- Blizzard code to set the EM set displayed in the EM Set dialog.
--
local function equipSetSpecRecalculate(setName, setIcon)
  -- Local Variables
  local level = UnitLevel("player");

  -- Check if the EM Set spec feature is enabled
  if (NigglesDBC.settings.UiEMSetSpec) then
    -- Work out which spec, if any, is selected for the EM set
    if (GearManagerDialogPopup.origName ~= nil) then
      equipSetSpecInfo.selectedSpec =
        (NigglesDBC.equipSetSpec[GearManagerDialogPopup.origName] or 0);
    else
      equipSetSpecInfo.selectedSpec = 0;
    end

    -- Initialise all the buttons
    for _, current in pairs(equipSetSpecInfo.buttons) do
      current:SetEnabled(level >= 10);
      current:SetChecked((level >= 10) and
        (equipSetSpecInfo.selectedSpec == current:GetID()));
    end
  end

  return;
end

--
-- Function to scan the tooltip for the specified item to get the maximum
-- character level it will scale to.
--
local function getHeirloomMaxLevel(itemId)
  -- Local Variables
  local numLines;
  local maxLevel = 0;

  -- Create the tooltip, if required
  if (NigglesEquippedTooltip == nil) then
    CreateFrame("GameTooltip", "NigglesEquippedTooltip",
      UIParent, "GameTooltipTemplate");
    NigglesEquippedTooltip:SetOwner(UIParent, "ANCHOR_NONE")
  end

  -- Set the tooltip
  NigglesEquippedTooltip:ClearLines();
  NigglesEquippedTooltip:SetItemByID(itemId);

  -- Scan the lines of the tooltip for the level range
  numLines = NigglesEquippedTooltip:NumLines();
  for lineIdx = numLines, 1, -1 do
    _, _, maxLevel = string.find(
      _G["NigglesEquippedTooltipTextLeft"..lineIdx]:GetText(),
      L["HeirloomLevelPattern"]);
    if (maxLevel ~= nil) then
      break;
    end
  end

  return (maxLevel or 80);
end

-------------------------------------------------------------------------------
  --               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to check if Pet Battle items aren't equipped at start of a battle.
--
-- NOTE: This function will need enhancement if Blizzard ever add items for
--       ring, trinket, main or off-hand slots.
--
function NigglesEquipmentCheckPetBattle()
  -- Local Variables
  local bestUnequipped = {};
  local itemLink;
  local itemTexture;
  local numUnequipped = 0;
  local string = "";
  local nonMaxLevel = false;

  -- Work out if there are any pets that aren't max level
  for idx = 1, C_PetBattles.GetNumPets(1) do
    if (C_PetBattles.GetLevel(1, idx) < MAX_PET_LEVEL) then
      nonMaxLevel = true;
      break;
    end
  end

  -- Check if there are any pets that aren't max level
  if (nonMaxLevel) then
    -- Initialise the best unequipped table
    for _, itemInfo in pairs(petBattleItems) do
      bestUnequipped[itemInfo.slot] = {bonus = 0, id = 0};
    end

    -- Check if the best Pet Battle equipment owned is equipped
    for _, itemInfo in ipairs(petBattleItems) do
      if ((itemInfo.bonus > bestUnequipped[itemInfo.slot].bonus) and
          (not IsEquippedItem(itemInfo.id)) and
          ((GetItemCount(itemInfo.id, false) > 0) or
           (GetItemCount(itemInfo.id, true) > 0))) then
        if (bestUnequipped[itemInfo.slot].id == 0) then
          numUnequipped = numUnequipped+1;
        end
        bestUnequipped[itemInfo.slot].id    = itemInfo.id;
        bestUnequipped[itemInfo.slot].bonus = itemInfo.bonus;
      end
    end

    -- Check if there are any items that should be equipped
    if (numUnequipped > 0) then
      -- Create the Niggles warning popup, if required
      if (Niggles.warningPopup == nil) then
        CreateFrame("Frame", "NigglesWarningPopup", UIParent,
          "NigglesWarningPopupTemplate");
      end

      -- Create the string for the unequipped items
      for _, unequippedInfo in pairs(bestUnequipped) do
        if (unequippedInfo.id ~= 0) then
          _, itemLink, _, _, _, _, _, _, _, itemTexture =
            GetItemInfo(unequippedInfo.id);
          -- NOTE: The '|r' is done after the '\n' to overcome a Blizzard bug
          string  = string.."\124T"..itemTexture..":0:0:0:-8\124t "..
            itemLink.."\n|r";
        end
      end

      -- Set the info to be displayed
      NigglesWarningPopup.text:SetText(L["PetBattleUnequipped"]);
      NigglesWarningPopup.list:SetText(string);
      NigglesWarningPopup.list:Show();
      NigglesWarningPopup.item:Hide();

      -- Display the frame
      StaticPopupSpecial_Show(NigglesWarningPopup);
    end
  end

  return;
end

--
-- Function to get the number of empty slots
--
function NigglesEquipmentGetEmptySlots()
  return equippedStats.empty;
end

--
-- Function to get the number of low level items equipped
--
function NigglesEquipmentGetLowIlvl()
  return equippedStats.lowIlvl;
end

--
-- Function to check if a lance is equipped
--
function NigglesEquipmentIsLance(itemId)
  return (tournamentLances[itemId] ~= nil);
end

--
-- Function to scan all equipped items. It works out an average ilvl, ignoring
-- active heirlooms, as well as counting the number of empty slots.
--
function NigglesEquipmentScan()
  -- Local Variables
  local averageIlvl = 0;
  local itemIlvl;
  local itemId;
  local itemQuality;
  local itemSubCat;
  local itemType;
  local lowIlvl;
  local infoAvailable = true;

  -- Reset the stats for the equipped items
  equippedStats.count   = 0;
  equippedStats.lowIlvl = 0;
  equippedStats.empty   = 0;

  -- Scan all equipment slots
  for slotId, slotInfo in pairs(equippedItems) do
    itemId = GetInventoryItemID("player", slotId);
    if (itemId == nil) then
      slotInfo.id         = 0;
      slotInfo.iLvl       = 0;
      equippedStats.empty = equippedStats.empty+1;
    else
      -- Check if the item has changed since the last scan
      if (itemId ~= slotInfo.id) then
        -- Get the item's info
        _, _, itemQuality, itemIlvl = GetItemInfo(itemId);
        if (itemQuality ~= nil) then
          slotInfo.id   = itemId;
          slotInfo.iLvl = itemIlvl;

          -- Check if the item is an active heirloom, giving two level of grace
          if ((itemQuality == 7) and
              (getHeirloomMaxLevel(itemId)+2 >= UnitLevel("player"))) then
            slotInfo.iLvl = ILVL_ACTIVEHEIRLOOM;
          end
        else
          slotInfo.id   = 0;
          slotInfo.iLvl = 0;
          infoAvailable = false;
          break;
        end
      end

      -- Update the info for calculating the average item level
      if ((infoAvailable) and
          (slotInfo.iLvl ~= ILVL_ACTIVEHEIRLOOM)) then
        averageIlvl         = averageIlvl+slotInfo.iLvl;
        equippedStats.count = equippedStats.count+1;
      end
    end
  end

  -- Check item info could be obtained for all items
  if (infoAvailable) then
    -- Adjust the number of empty slots to account for 2H weapons
    if ((equippedItems[INVSLOT_MAINHAND].id ~= 0) and
        (equippedItems[INVSLOT_OFFHAND].id  == 0)) then
      _, _, _, _, _, _, itemSubCat, _, itemType =
        GetItemInfo(equippedItems[INVSLOT_MAINHAND].id);
      if ((itemType == "INVTYPE_2HWEAPON") or
          (equippedTwoHanded[itemSubCat] ~= nil)) then
        equippedStats.empty = equippedStats.empty-1;
      end
    end

    -- Work out the average item level
    if (equippedStats.count > 0) then
      averageIlvl = averageIlvl/equippedStats.count;
    else
      averageIlvl = 0;
    end

    -- Work out how many low level items are equipped
    lowIlvl = averageIlvl*LOWLEVEL_MULTIPLIER;
    for _, slotInfo in pairs(equippedItems) do
      if ((slotInfo.id   > 0) and
          (slotInfo.iLvl ~= ILVL_ACTIVEHEIRLOOM) and
          (slotInfo.iLvl < lowIlvl)) then
        equippedStats.lowIlvl = equippedStats.lowIlvl+1;
      end
    end
  else
    -- Reset the stats for the equipped items
    equippedStats.count   = 0;
    equippedStats.lowIlvl = 0;
    equippedStats.empty   = 0;

    -- Create a timer to scan the equipment again
    NigglesTimerCreate("EquipmentScan", NigglesEquipmentScan,
      nil, nil, 1.0, false);
  end

  -- Update the low level item indicators
  if ((NigglesDBC.settings.UiLowLevelItems) and
      (equippedStats.lowIlvl > 0)) then
    lowIlvl = averageIlvl*LOWLEVEL_MULTIPLIER;
    for slotId, slotInfo in pairs(equippedItems) do
      if ((slotInfo.id   > 0) and
          (slotInfo.iLvl ~= ILVL_ACTIVEHEIRLOOM) and
          (slotInfo.iLvl < lowIlvl)) then
        _G["NigglesLowIlvl"..slotId]:Show();
      else
        _G["NigglesLowIlvl"..slotId]:Hide();
      end
    end
  else
    for slotId, _ in pairs(equippedItems) do
      _G["NigglesLowIlvl"..slotId]:Hide();
    end
  end

  return;
end

--
-- Function to initialise the Niggles Equipment functions
--
function NigglesEquipmentInit()
  -- Create the frames used to indicate low level items
  -- NOTE: A font string can't simply be created as a child of the button
  --       directly because it can't be placed above the cooldown.
  for slotId, slotName in pairs(equipSlotNames) do
    CreateFrame("Frame", "NigglesLowIlvl"..slotId,
      _G["Character"..slotName.."Slot"], "NigglesLowIlvlTemplate");
  end

  -- Hook the function used to sell items
  hooksecurefunc("UseContainerItem", NigglesEquipmentOnItemSale);

  -- Add the 'Equip Set' popup to the list of frames closed by ESC
  tinsert(UISpecialFrames, "NigglesEquipSetPopup");

  return;
end

--
-- Function called when an item is being sold
--
function NigglesEquipmentOnItemSale(container, slot, target)
  -- Local Variables
  local popup;

  -- Check if the item is being sold
  if ((MerchantFrame:IsShown()) and
      (MerchantFrame.selectedTab == 1)) then
    -- Check if the item is part of a set and a warning should be displayed
    if ((NigglesDBC.settings.WarningEquipSetItem) and
        (GetContainerItemEquipmentSetInfo(container, slot))) then
      local itemTexture, itemCount, _, itemQuality, _, _, itemLink =
        GetContainerItemInfo(container, slot);
      local _, _, itemName = string.find(itemLink, "|h%[(.*)%]|h|r$");

      -- Create the Niggles warning popup, if required
      if (NigglesWarningPopup == nil) then
        CreateFrame("Frame", "NigglesWarningPopup", UIParent,
          "NigglesWarningPopupTemplate");
      end
      popup = NigglesWarningPopup;

      -- Set the info to be displayed
      popup.text:SetText(L["EquipSetItemSold"]);
      popup.list:Hide();
      popup.item.container = container;
      popup.item.slot      = slot;
      popup.item.icon:SetTexture(itemTexture);
      popup.item.text:SetTextColor(
        ITEM_QUALITY_COLORS[itemQuality].r,
        ITEM_QUALITY_COLORS[itemQuality].g,
        ITEM_QUALITY_COLORS[itemQuality].b,
        1);
      popup.item.text:SetText(itemName);
      popup.item:Show();

      -- Display the frame
      StaticPopupSpecial_Show(NigglesWarningPopup);
    end
  end

  return;
end

--
-- Function called when an EM set is equipped successfully.
--
function NigglesEquipSetEquipped(setName)
  -- Check if a spec exists for the EM set & not in combat
  if ((setName ~= nil) and
      (NigglesDBC.equipSetSpec[setName] ~= nil) and
      (not InCombatLockdown())) then
    -- Local Variables
    local setSpecId = NigglesDBC.equipSetSpec[setName];
    local setSpecIdx = 0;
    local specIdx = 0;

    -- Search the available specialisations for the set's
    for idx = 1, GetNumSpecGroups() do
      specIdx = GetSpecialization(nil, false, idx);
      if ((specIdx ~= nil) and 
          (GetSpecializationInfo(specIdx) == setSpecId)) then
        setSpecIdx = idx;
        break;
      end
    end

    -- Check the spec is available, isn't active and it's safe to change
    -- NOTE: Druids can change spec while in flight form
    if ((setSpecIdx > 0) and
        (setSpecIdx ~= GetActiveSpecGroup()) and
        ((not IsFlying()) or (not IsMounted()))) then
      -- Ensure the spec change doesn't trigger an EM set change
      NigglesEquipSetPopup.processEvent = false;
      NigglesEquipSetPopup:RegisterUnitEvent("UNIT_SPELLCAST_INTERRUPTED",
        "player");

      -- Set the active spec
      SetActiveSpecGroup(setSpecIdx);
    end
  end

  return;
end

--
-- Function called when the Niggles 'Equip Set' popup is loaded
--
function NigglesEquipSetPopupOnLoad(self)
  -- Local Variables
  local scrollFrame = self.scrollFrame

  -- Initialise some variables
  self.processEvent = true;
  self.setName      = nil;

  -- Set the frame's script handlers
  self:SetScript("OnShow", equipSetPopupOnShow);
  self:SetScript("OnEvent", equipSetPopupOnEvent);

  -- Register the frame for dragging
  self:RegisterForDrag("LeftButton");
  
  -- Initialise the scroll frame
  scrollFrame.update = NigglesEquipSetPopupUpdate;
  scrollFrame.scrollBar.doNotHide = 1;
  HybridScrollFrame_OnLoad(scrollFrame);
  HybridScrollFrame_CreateButtons(scrollFrame, "NigglesEquipSetBtnTemplate");

  -- Work around a problem with the first button shifting left-right
  scrollFrame.buttons[1]:SetWidth(scrollFrame.buttons[1]:GetWidth()-1);

  return;
end

--
-- Function to update the Niggles 'Equip Set' popup
--
function NigglesEquipSetPopupUpdate()
  -- Local Variables
  local scrollFrame = NigglesEquipSetPopupScrollFrame;
  local parent = NigglesEquipSetPopup;
  local numFiltered = filteredEquipSets[0];
  local scrollOffset = HybridScrollFrame_GetOffset(scrollFrame);
  local buttons = scrollFrame.buttons;
  local selectedSet = parent.selectedSet;
  local setName;
  local setIcon;
  local button;
  local numLost;

  -- Update the scroll frame's buttons
  for buttonIdx = 1, #buttons do
    -- Assign the button to a more convenient variable
    button = buttons[buttonIdx];

    -- Check if there is an equipment set for the button to display
    if (buttonIdx+scrollOffset <= numFiltered) then
      -- Get the equipment set info
      setName, setIcon, _, isEquipped, _, _, _, numLost =
        GetEquipmentSetInfo(filteredEquipSets[buttonIdx+scrollOffset]);

      -- Set the name of the equipment set the button is for
      button.setName = setName;
      button.text:SetText(setName);

      -- Set the colour of the button's text
      if (numLost > 0) then
        button.text:SetTextColor(RED_FONT_COLOR.r, RED_FONT_COLOR.g,
          RED_FONT_COLOR.b);
      else
        button.text:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g,
          NORMAL_FONT_COLOR.b);
      end

      -- Set the button's textures
      button.icon:SetTexture(setIcon or
        "Interface\\Icons\\INV_Misc_QuestionMark");
      button.selected:SetShown(
        (selectedSet ~= nil) and (button.setName == selectedSet));
      button.equipped:SetShown(isEquipped);
      if ((buttonIdx+scrollOffset) == 1) then
        button.bgTop:Show();
        button.bgMiddle:SetPoint("TOP", button.bgTop, "BOTTOM");
      else
        button.bgTop:Hide();
        button.bgMiddle:SetPoint("TOP", button, "TOP", 0, 0);
      end
      if ((buttonIdx+scrollOffset) == numFiltered) then
        button.bgBottom:Show();
        button.bgMiddle:SetPoint("BOTTOM", button.bgBottom, "TOP");
      else
        button.bgBottom:Hide();
        button.bgMiddle:SetPoint("BOTTOM");
      end
      button.stripe:SetShown((buttonIdx+scrollOffset)%2 == 0);
      
      -- Show the button
      button:Show();
    else
      button:Hide();
    end
  end

  -- Update the scroll frame
  HybridScrollFrame_Update(scrollFrame, numFiltered*buttons[1]:GetHeight()+6,
    scrollFrame:GetHeight());
  
  return;
end

--
-- Function called when a button in the Niggles 'Equip Set' scroll frame is
-- clicked.
--
function NigglesEquipSetBtnOnClick(self, button, down)
  -- Check the button has a set name associated with it
  if ((self.setName ~= nil) and (self.setName ~= "")) then
    -- Save the name of the selected set
    NigglesEquipSetPopup.selectedSet = self.setName;
    NigglesEquipSetPopup.equip:Enable();

    -- Update the scroll frame
    NigglesEquipSetPopupUpdate();
  end

  return;
end

--
-- Function called when a button in the Niggles 'Equip Set' scroll frame is
-- double-clicked.
--
function NigglesEquipSetBtnOnDoubleClick(self, button, down)
  -- Check the button has a set name associated with it
  if ((self.setName ~= nil) and (self.setName ~= "")) then
    -- Equip the button's equipment set
    PlaySound("igCharacterInfoTab");
    EquipmentManager_EquipSet(self.setName);

    -- Hide the 'Equip Set' popup
    NigglesEquipSetPopup:Hide();
  end

  return;
end

--
-- Function called when the mouse enters a button in the Niggles 'Equip Set'
-- scroll frame.
--
function NigglesEquipSetBtnOnEnter(self)
  -- Check the button has a set name associated with it
  if ((self.setName ~= nil) and (self.setName ~= "")) then
    GameTooltip_SetDefaultAnchor(GameTooltip, self);
    GameTooltip:SetEquipmentSet(self.setName);
  end

  return;
end

--
-- Function called when the 'Equip' button is clicked in the Niggles 'Equip
-- Set' popup
--
function NigglesEquipSetEquipOnClick(self)
  -- Check there is a selected set
  if ((NigglesEquipSetPopup.selectedSet ~= nil) and 
      (NigglesEquipSetPopup.selectedSet ~= "")) then
    -- Equip the selected equipment set
    PlaySound("igCharacterInfoTab");
    EquipmentManager_EquipSet(NigglesEquipSetPopup.selectedSet);

    -- Hide the 'Equip Set' popup
    NigglesEquipSetPopup:Hide();
  end

  return;
end

--
-- Function  to process a click on an equipment set spec button. It makes sure
-- that the button that was clicked is checked and that it is the only button
-- that is checked.
--
function NigglesEquipSetSpecBtnOnClick(self, mouseBtn)
  -- Save the new selected spec
  equipSetSpecInfo.selectedSpec = self:GetID();

  -- Un-check all other button
  for _, current in pairs(equipSetSpecInfo.buttons) do
    if (current ~= self) then
      current:SetChecked(false);
    end
  end

  -- Make sure the clicked button is checked
  if (not self:GetChecked()) then
    self:SetChecked(true);
  end

  return;
end

--
-- Function to display the tooltip for an EM set spec button
--
function NigglesEquipSetSpecBtnOnEnter(self)
  -- Local Variables
  local specId = self:GetID();
  local specName = nil;

  -- Get the name of the button's specialization
  if (specId > 0) then
    _, specName = GetSpecializationInfoByID(specId);
  else
    specName = NONE;
  end

  -- Display a tooltip for the button
  GameTooltip:SetOwner(self, "ANCHOR_RIGHT", 0);
  GameTooltip:SetText(specName,
    HIGHLIGHT_FONT_COLOR.r,
    HIGHLIGHT_FONT_COLOR.g,
    HIGHLIGHT_FONT_COLOR.b);
  GameTooltip:Show();

  return;
end

--
-- Function to enable/disable the EM Set spec functionality
--
function NigglesEquipSetSpecEnable()
  -- Local Variables
  local dialog = GearManagerDialogPopup;
  local scrollFrame = GearManagerDialogPopupScrollFrame;
  local child;
  local heightAdj = 70;
  local numSpecs = GetNumSpecializations();
  local specId = 0;
  local specTexture =
          "Interface\\PaperDollInfoFrame\\UI-GearManager-LeaveItem-Opaque";

  -- Check if the EM Set spec feature is enabled
  if ((NigglesDBC.settings.UiEMSetSpec) and
      (equipSetSpecInfo.textures[1] == nil)) then
    -- Create textures to cater for the increased size of the dialog
    child = dialog:CreateTexture(nil, "BACKGROUND", nil, 1);
    child:SetTexture("Interface\\MacroFrame\\MacroPopup-TopLeft");
    child:SetWidth(256);
    child:SetHeight(heightAdj);
    child:SetTexCoord(0, 1, 72/256, (72+heightAdj)/256);
    child:SetPoint("BOTTOMLEFT", dialog, "BOTTOMLEFT", 0, 42);
    equipSetSpecInfo.textures[1] = child;
    child = dialog:CreateTexture(nil, "BACKGROUND", nil, 1);
    child:SetTexture("Interface\\MacroFrame\\MacroPopup-TopRight");
    child:SetWidth(64);
    child:SetHeight(heightAdj);
    child:SetTexCoord(0, 1, 72/256, (72+heightAdj)/256);
    child:SetPoint("BOTTOMRIGHT", dialog, "BOTTOMRIGHT", 23, 42);
    equipSetSpecInfo.textures[2] = child;

    -- Create textures for a divider below the new section in the dialog
    child = dialog:CreateTexture(nil, "BACKGROUND", nil, 1);
    child:SetTexture("Interface\\MacroFrame\\MacroPopup-TopLeft");
    child:SetWidth(8);
    child:SetHeight(16);
    child:SetTexCoord(0.02734375, 0.05859375, 0.21484375, 0.27734375);
    child:SetPoint("BOTTOMLEFT", scrollFrame, "TOPLEFT", 45, -4);
    equipSetSpecInfo.textures[3] = child;
    child = dialog:CreateTexture(nil, "BACKGROUND", nil, 1);
    child:SetTexture("Interface\\MacroFrame\\MacroPopup-TopRight");
    child:SetWidth(8);
    child:SetHeight(16);
    child:SetTexCoord(0.390625, 0.515625, 0.21484375, 0.27734375);
    child:SetPoint("BOTTOMRIGHT", scrollFrame, "TOPRIGHT", 31, -4);
    equipSetSpecInfo.textures[4] = child;
    child = dialog:CreateTexture(nil, "BACKGROUND", nil, 1);
    child:SetTexture("Interface\\MacroFrame\\MacroPopup-TopLeft");
    child:SetHeight(8);
    child:SetTexCoord(0.5859375, 1, 0.23046875, 0.26171875);
    child:SetPoint("TOPLEFT", equipSetSpecInfo.textures[3], "TOPRIGHT", 0, -4);
    child:SetPoint("TOPRIGHT", equipSetSpecInfo.textures[4], "TOPLEFT", 0, -4);
    equipSetSpecInfo.textures[5] = child;

    -- Create a title for the new section
    child = dialog:CreateFontString();
    child:SetFontObject("GameFontHighlightSmall");
    child:SetText("Choose a Specialization:");
    child:SetPoint("BOTTOMLEFT", scrollFrame, "TOPLEFT", 62, heightAdj-12);
    equipSetSpecInfo.fontStrings[1] = child;

    -- Create the buttons for selecting a EM set's spec
    for idx = 0, numSpecs do
      -- Get the info for the spec the button is for
      if (idx > 0) then
        specId, _, _, specTexture = GetSpecializationInfo(idx);
      end

      -- Create the button
      child = CreateFrame("CHECKBUTTON", nil, dialog,
        "NigglesEquipSetSpecBtnTemplate", idx);
      childWidth = child:GetWidth();
      child:SetPoint("LEFT", dialog, "CENTER",
        (idx*(childWidth+5))-((((numSpecs+1)*childWidth)+(numSpecs*5))/2), 0);
      child:SetPoint("TOP", equipSetSpecInfo.fontStrings[1], "BOTTOM", 0, -6);
      child:SetID(specId);
      child:SetNormalTexture(specTexture);
      equipSetSpecInfo.buttons[idx] = child;
    end

    -- Hook the dialog's event handlers
    hooksecurefunc("RecalculateGearManagerDialogPopup",
      equipSetSpecRecalculate);
    GearManagerDialogPopupOkay:HookScript("PreClick", equipSetSpecOkayPreClick);
    GearManagerDialogPopupOkay:HookScript("OnClick",  equipSetSpecOkayOnClick);
  end

  -- Check if the EM set/spec feature is enabled
  if (NigglesDBC.settings.UiEMSetSpec) then
    -- Adjust the height of the dialog
    dialog:SetHeight(equipSetSpecInfo.dialogHeight+
     (NigglesDBC.settings.UiEMSetSpec and heightAdj or 0));

    -- Show/Hide the frames
    for _, current in pairs(equipSetSpecInfo.textures) do
      current:SetShown(NigglesDBC.settings.UiEMSetSpec);
    end
    for _, current in pairs(equipSetSpecInfo.fontStrings) do
      current:SetShown(NigglesDBC.settings.UiEMSetSpec);
    end
    for _, current in pairs(equipSetSpecInfo.buttons) do
      if (current ~= nil) then
        current:SetShown(NigglesDBC.settings.UiEMSetSpec);
      end
    end
  end

  -- (Un)register for events
  if (NigglesDBC.settings.UiEMSetSpec) then
    NigglesEquipSetPopup:RegisterEvent("EQUIPMENT_SWAP_FINISHED");
    NigglesEquipSetPopup:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED");
    NigglesEquipSetPopup:RegisterEvent("EQUIPMENT_SETS_CHANGED");
  else
    NigglesEquipSetPopup:UnregisterEvent("EQUIPMENT_SWAP_FINISHED");
    NigglesEquipSetPopup:UnregisterEvent("ACTIVE_TALENT_GROUP_CHANGED");
    NigglesEquipSetPopup:UnregisterEvent("EQUIPMENT_SETS_CHANGED");
  end
  if ((NigglesDBC.settings.UiLowLevelItems) or
      (NigglesDBC.settings.WarningLowLevelItems) or
      (NigglesDBC.settings.UiEMSetSpec)) then
    NigglesEquipSetPopup:RegisterEvent("PLAYER_EQUIPMENT_CHANGED");
  else
    NigglesEquipSetPopup:UnregisterEvent("PLAYER_EQUIPMENT_CHANGED");
  end
  
  -- Scan the equipped items, if required
  if ((NigglesDBC.settings.UiLowLevelItems) or
      (NigglesDBC.settings.WarningLowLevelItems)) then
    NigglesEquipmentScan();
  end

  return;
end

--
-- Function to initialise the EM set spec. It adjusts the points on various
-- frames in the EM Set panel to better accommodate additional frames.
--
function NigglesEquipSetSpecInit()
  -- Local Variables
  local dialog = GearManagerDialogPopup;
  local scrollFrame = GearManagerDialogPopupScrollFrame;
  local iconButton = GearManagerDialogPopupButton1;
  local child;
  local childType;
  local childText;
  local xOffset;
  local yOffset;
  local setId;

  -- Remove info for any EM sets that no longer exist
  for setName, _ in pairs(NigglesDBC.equipSetSpec) do
    _, setId = GetEquipmentSetInfoByName(setName);
    if (setId == nil) then
      NigglesDBC.equipSetSpec[setName] = nil;
    end
  end

  -- Save the original height of the dialog
  equipSetSpecInfo.dialogHeight = dialog:GetHeight();

  -- Reposition the dialog's unnamed textures and font strings
  for _, child in ipairs({dialog:GetRegions()}) do
    childType = child:GetObjectType();
    if (childType == "Texture") then
      _, _, _, xOffset, yOffset = child:GetPoint("TOPLEFT");
      if ((xOffset == 0) and (yOffset == 0)) then
        child:ClearAllPoints();
        child:SetPoint("TOPLEFT", dialog, "TOPLEFT", 0, 0);
      elseif ((xOffset > 0) and (yOffset == 0)) then
        child:ClearAllPoints();
        child:SetPoint("TOPRIGHT", dialog, "TOPRIGHT", 23, 0);
      elseif ((xOffset == 0) and (yOffset < 0)) then
        child:ClearAllPoints();
        child:SetPoint("BOTTOMLEFT", dialog, "BOTTOMLEFT", 0, -22);
      elseif ((xOffset > 0) and (yOffset < 0)) then
        child:ClearAllPoints();
        child:SetPoint("BOTTOMRIGHT", dialog, "BOTTOMRIGHT", 23, -22);
      end
    elseif (childType == "FontString") then
      childText = child:GetText();
      if (childText == MACRO_POPUP_CHOOSE_ICON) then
        child:ClearAllPoints();
        child:SetPoint("TOPLEFT", scrollFrame, "TOPLEFT", 62, -1);
      end
    end
  end

  -- Reposition the scroll frame
  scrollFrame:ClearAllPoints();
  scrollFrame:SetPoint("BOTTOMLEFT", dialog, "BOTTOMLEFT", -38, 35);

  -- Reposition all the icon buttons
  iconButton:ClearAllPoints();
  iconButton:SetPoint("TOPLEFT", scrollFrame, "TOPLEFT", 62, -17);

  return;
end

--
-- Function called when the active spec changes. It counts how many equipment
-- sets are for the new active spec. If there is one, that set is automatically
-- equipped. If there's more than one then a popup is displayed for the user
-- to select which set to equip.
--
function NigglesEquipSetSpecGroupChanged()
  -- Local Variables
  local lastMatch = nil;
  local setName;
  local setSpecId
  local activeSpecId;
  local activeSpecIdx = GetSpecialization();
  local equipSetSpec = NigglesDBC.equipSetSpec;
  local numSets = 0;

  -- Check there is an active spec
  if (activeSpecIdx ~= nil) then
    -- Get the ID of the active spec
    activeSpecId = GetSpecializationInfo(activeSpecIdx);

    -- Count the number of equipment sets for the active spec
    if (activeSpecId ~= nil) then
      for idx = 1, GetNumEquipmentSets() do
        setName = GetEquipmentSetInfo(idx);
        if (equipSetSpec[setName] == activeSpecId) then
          numSets   = numSets+1;
          lastMatch = setName;
        end
      end
    end

    -- Check how many sets were found for the active spec
    if (numSets == 1) then
      UseEquipmentSet(lastMatch);
    elseif (numSets > 1) then
      -- Show the 'Equip Set' popup
      NigglesEquipSetPopup:Show();
    end
  end

  return;
end

